

#
# These are used for nicely wrapping the comments of classes, variables, etc.
#
from textwrap import TextWrapper

tw = TextWrapper()
tw50 = TextWrapper( width = 50 )
def wrap_h( text ):
    """wrap using the default 70 character wrap limit, for all lines
    then join with newlines and indentation by 18 spaces"""
    return ('\n'+' '*18).join( tw.wrap( text ) )
def wrap_cpp( text ):
    return ('\n'+' '*14).join( tw.wrap( text ) )
def wrap_param_h( text ):
    """wrap using 50 character wrap limit, for all lines
    then join with newlines and indentation by 35 spaces"""
    return ('\n'+' '*35).join( tw50.wrap( text ) )
def wrap_param_cpp( text ):
    return ('\n'+' '*31).join( tw50.wrap( text ) )


class Arg :
    """Simple class for arguments, contains type, name, and comment"""
    def __init__( self, type, name, comment = '<NO DESCRIPTION>' ):
        """
        type should be int, int *, int &, and so on
        name may contain also a default value, or array length, e.g., blah, blah[3], blah = 1, and so on
        """
        self.type = type
        #self.fullname = name
        self.comment = comment

        # find about a default value
        if '=' in name:
            self.name = name.split('=')[0].strip()
            self.defval = name.split('=')[-1].strip()
        else:
            self.name = name
            self.defval = None

        # find about array length
        if '[' in self.name:
            self.basename = self.name.split('[')[0].strip()
        else:
            self.basename = self.name
        
    def decl( self ):
        """Used in header file"""
        #return self.type + ' ' + self.fullname
        if self.defval:
            return self.type + ' ' + self.name + ' = ' + self.defval
        return self.type + ' ' + self.name

    def corename( self ):
        return self.basename

    #def name( self ):
    #    # if there's a default value, get rid of it
    #    return self.fullname.split('=')[0].strip()

    #def default_value( self ):
    #    # if there's a default value, return it, otherwise None
    #    if '=' in self.fullname:
    #        return self.fullname.split('=')[-1].strip()
    #    else:
    #        return None

    #def defn( self, code ):
    #    """Used in code (cpp)"""
    #    # if there's a default value, get rid of it
    #    name = self.name()
    #    # check whether the variable is actually used
    #    if self.corename in code:
    #        # yes
    #        return self.type + ' ' + name
    #    else:
    #        # no, put it in comments to avoid warnings
    #        return self.type + ' /* ' + name + ' */'

    def defn( self, code ):
        """Used in code (cpp)"""
        # check whether the variable is actually used
        if self.basename in code:
            # yes
            return self.type + ' ' + self.name
        else:
            # no, put it in comments to avoid warnings
            return self.type + ' /* ' + self.name + ' */'

    #def doc( self, indent ):
    #    """A line for documentation string for header files, indented by indent spaces"""
    #    s = '\n' + ' ' * indent
    #    dv = self.default_value()
    #    if dv:
    #        # the argument has a default value, document it
    #        return s + '@param      %-15s  (Defaults to %s) %s' % (self.name(), dv, wrap_param_h( self.comment ))
    #    else:
    #        return s + '@param      %-15s  %s' % (self.fullname, wrap_param_h( self.comment ))

    def doc( self, indent ):
        """A line for documentation string for header files, indented by indent spaces"""
        s = '\n' + ' ' * indent
        if self.defval:
            # the argument has a default value, document it
            return s + '@param      %-15s  (Defaults to %s) %s' % (self.name, self.defval, wrap_param_h( self.comment ))
        else:
            return s + '@param      %-15s  %s' % (self.name, wrap_param_h( self.comment ))


class Enum :
    """Simple class for enums"""
    def __init__( self, name_and_doc, value_doc_pairs, var = '' ):
        # first two arguments are either a string or a tuple of two strings,
        # the second string being documentation
        if type( name_and_doc ) == type( '' ):
            self.name    = name_and_doc
            self.comment = ''
        else:
            assert type( name_and_doc ) == type( ('','') )
            self.name    = name_and_doc[0]
            self.comment = name_and_doc[1]
        # also with value_doc_pairs the doc is optional
        # the value can be either 'KBlah' or can have the value 'KBlah = 1', etc.
        self.value_doc_pairs = value_doc_pairs
        self.var             = var
        
    def declaration( self ):
        text = """\
    /*! @enum %(name)s
        @discussion %(comment)s%(valuedocs)s
    */
    enum %(name)s
        {
%(items)s
        }%(var)s;
"""
        self.valuedocs = ''
        for vd in self.value_doc_pairs:
            if type(vd) == type(''):
                # just the value, no doce
                self.valuedocs += '\n        @value %s' % vd.split()[0]
            else:
                # both value and document string
                self.valuedocs += '\n        @value %s %s' % (vd[0].split()[0], vd[1])
        #self.valuedocs = '\n'.join( ['        @value %s %s' % tuple(pair) for pair in self.value_doc_pairs] )
        self.items = ',\n'.join( ['        %s' % pair[0] for pair in self.value_doc_pairs] )
        if self.var:
            self.var = ' '+self.var
        return text % self.__dict__

    def definition( self ):
        return ''

        
class Var :
    """Simple class for variables, contains type, name, and comment"""
    def __init__( self, type, name, comment = '<NO DESCRIPTION>' ):
        self.type = type
        self.name = name
        self.comment = comment

    def declaration( self ):
        text = """\
    /*! @var %(name)s %(comment)s */
    %(type)s %(name)s;

"""
        return text % self.__dict__

    def definition( self ):
        return ''

        
class Method :
    decl_text = """\
    /*!
      @function   %(name)s
      @discussion %(discussion_formatted)s%(params)s%(resdoc_formatted)s
    */
    %(RESULT)s%(name)s( %(ARGS_decl)s )%(CONST)s;

"""
    def_text = """\

/*
  @function   %(name)s
  @discussion %(discussion_formatted)s%(params)s%(resdoc_formatted)s
*/
%(RESULT)s$CLASS$::%(name)s( %(ARGS_def)s )%(CONST)s %(CONSTRUCT_INIT)s
{    %(code)s
}
"""
    def __init__( self, result_and_doc, name_and_discussion, args, code,
                  construct_init = [],     # only for constructors, member initializations, a string or list of strings
                  is_const = False,        # is it a constant method
                  is_virtual = False,      # is it a virtual method
                  is_pure_virtual = False):# is it a pure virtual method
        # the first two arguments are either a string or a tuple of two strings,
        # the second string being documentation
        if type( result_and_doc ) == type( '' ):
            self.result = result_and_doc
            if self.result == 'void' or self.result == '':
                self.resdoc = ''
            else:
                self.resdoc = '<NO DESCRIPTION>'
        else:
            assert type( result_and_doc ) == type( ('','') )
            self.result = result_and_doc[0]
            self.resdoc = result_and_doc[1]
        # the same thing for name and discussion (either just name or a tuple of name and doc)
        if type( name_and_discussion ) == type( '' ):
            self.name       = name_and_discussion
            self.discussion = '<NO DESCRIPTION>'
        else:
            assert type( name_and_discussion ) == type( ('','') )
            self.name       = name_and_discussion[0]
            self.discussion = name_and_discussion[1]
        self.args           = args[:]
        self.code           = code
        # the input could be string or list, make it be a list
        if type(construct_init) == type([]):
            self.construct_init = construct_init
        else:
            assert type(construct_init) == type('')
            if construct_init == '':
                # in case passing an emptry string, make it an empty list
                self.construct_init = [ ]
            else:
                # there's a real string, put it into list
                self.construct_init = [ construct_init ]
        self.CONST = ''
        if is_const:
            self.CONST = ' const'
        self.is_virtual = is_virtual
        self.is_pure_virtual = is_pure_virtual

    def declaration( self ):
        """Create a declaration for the header file"""
        self.RESULT = self.result
        # if there's a result type, add a separating space
        if self.RESULT: self.RESULT += ' '
        if self.is_pure_virtual:
            self.CONST += ' = 0'
        if self.is_virtual or self.is_pure_virtual:
            self.RESULT = 'virtual ' + self.RESULT
        if self.is_dll and not self.is_pure_virtual:  # is_dll must be added by CClass!
            self.RESULT = 'IMPORT_C ' + self.RESULT
        # format documentation
        self.discussion_formatted = wrap_h( self.discussion )
        self.params = ''.join( [x.doc(6) for x in self.args] )
        if self.resdoc:
            self.resdoc_formatted = '\n      @result     ' + wrap_h( self.resdoc )
        else:
            self.resdoc_formatted = ''
        # format arguments (or void if no args)
        self.ARGS_decl = ', '.join( [x.decl() for x in self.args] )
        if not self.ARGS_decl:
            self.ARGS_decl = 'void'
        return self.decl_text % self.__dict__

    def definition( self ):
        """Create a definition for the source file"""
        if self.is_pure_virtual:
            return ''
        # remove static from return type if there
        self.RESULT = self.result.replace( 'static', '' ).strip()
        # if there's a result type, add a separating space
        if self.RESULT: self.RESULT = self.RESULT + ' '
        if self.is_dll:  # this must be added by CClass!
            self.RESULT = 'EXPORT_C ' + self.RESULT
        if self.construct_init:
            self.CONSTRUCT_INIT = '\n: ' + ', '.join( self.construct_init )
        else:
            self.CONSTRUCT_INIT = ''
        if self.code:
            self.code = '\n' + self.indent( self.code.strip() )
        # if the argument is not used in the function, comment it out to avoid warnings (or void if no args)
        self.ARGS_def = ', '.join( [x.defn( self.code + self.CONSTRUCT_INIT ) for x in self.args] )
        if not self.ARGS_def:
            self.ARGS_def = 'void'
        # format documentation
        self.discussion_formatted = wrap_cpp( self.discussion )
        self.params = ''.join( [x.doc(2) for x in self.args] )
        if self.resdoc:
            self.resdoc_formatted = '\n  @result     ' + wrap_cpp( self.resdoc )
        else:
            self.resdoc_formatted = ''
        return self.def_text % self.__dict__

    def indent( self, code, begin_indent = 4, step = 4 ):
        """Indent the code following the following simple rule:
        at start the code is indented by begin_indent, if the first character of the
        row is { increase indentation by step, if } decrease by step"""
        lines = []
        i = begin_indent
        for line in [line.strip() for line in code.split('\n')]:
            if line:
                if line[0] == '}':
                    i -= step
                lines.append( (' ' * i) + line )
                if line[0] == '{':
                    i += step
            else:
                lines.append( '' )
        return '\n'.join( lines )
                
