"""
symwiz
A wizard for creating Symbian projects

Kari Pulli
kari.pulli@nokia.com
"""

import os, sys, shutil, re
from os.path import join
from optparse import OptionParser
import difflib

import preferences # user preferences for use of symwiz

from method import *

ENG_UID = ''
EngName = ''

def use_engine( eng_name, eng_uid, document_name = '' ):
    """If you want to separate the application engine to a dll, and just do the UI
    using the Symbian framework, call this function"""
    global ENG_UID, EngName, engine
    EngName = eng_name
    ENG_UID = eng_uid

    engine = CClass( (EngName, 'An instance of this class is the entry to the dll engine') )
    engine.parent = ' : public CBase'
    engine.is_dll = True
    engine.two_phase( [], '', '', '')
    engine.add_public( Method( 'void', ('Clear', 'Clear the engine / model'), [], '' ) )
    
    application.intro_src += '\nconst TUid KUid$PrjName$Ui  = {$ENG_UID$};'
    application.add_public( Method( 'TUid', 'AppUiUid', [], """    return KUid$PrjName$Ui;""", is_const = True) )
    document.inc_inc += [ '"$EngName$.h"' ]
    document.add_private( Var('C$EngName$*', 'iModel', 'The model'), destroy_var = True )
    document.add_public( Method( ('C$EngName$*', 'Returns a reference to the model' ),
                                 ('Model', 'Get the model associated with the document'), [], 
                                 '    return iModel;' ) )
    document.members['ConstructL'].code += '\n    iModel = C$EngName$::NewL();'
    appui.add_public( Method( 'void', ('HandleModelChangeL', 'Update the system to reflect changes in the model'),
                              [], """\
    iAppView->DrawNow();          // Redraw the application view based on the new model.
    iEikonEnv->UpdateTaskNameL(); // Update the file name in the task list to show the current file""" ) )
    appview.inc_src  += [ '"$EngName$.h"' ]

    if document_name:
        # add also facility to store data, to create persistence
        resource.default_document = document_name
        resource.libraries += [ 'estor' ]
        resource.eng_libs  += [ 'estor' ]
        document.add_public( Method( 'void',
                                     ('StoreL', 'Store the document to the store'),
                                     [Arg('CStreamStore&', 'aStore', 'the store to save the data to'),
                                      Arg('CStreamDictionary&', 'aStreamDict',
                                          'the dictionary to hold the data stream id') ], """\
    // Store model information first
    //  Get the model to save itself to the store
    TStreamId modelStreamId = iModel->StoreL(aStore);
    //  Add an entry into the dictionary for the model data
    aStreamDict.AssignL(Application()->AppDllUid(), modelStreamId);

    // Now get the Ui to store itself to the store
    TStreamId uiStreamId = static_cast<C$PrjName$AppUi*>(iAppUi)->StoreL(aStore);
    //  Add an entry for it to the dictionary
    aStreamDict.AssignL(static_cast<C$PrjName$Application*>(Application())->AppUiUid(), uiStreamId);""",
                                     is_const = True) )          
        document.add_public( Method( 'void',
                                     ('RestoreL', 'Restore the document from the store'),
                                     [Arg('const CStreamStore&', 'aStore', 'the store to restore the data from'),
                                      Arg('const CStreamDictionary&', 'aStreamDict',
                                          'the dictionary to hold the data stream id') ], """\
    // Restore the model from the appropriate store stream.
    // Do not set this as the document's model yet.
    // That way, if the restore fails, we have still have the old model.

    C$EngName$* newModel = C$EngName$::NewL();
    CleanupStack::PushL(newModel);

    //  Find the stream id for the model's data
    TStreamId modelStreamId = aStreamDict.At(Application()->AppDllUid());
    //  Restore the model from that stream 
    newModel->RestoreL(aStore, modelStreamId);

    //  Now get rid of the old model and set the new one as current
    delete iModel;
    iModel = newModel;
    CleanupStack::Pop(newModel);

    //  Just restored so can't have changed
    SetChanged(EFalse); 

    // Now restore UI data
    TStreamId uiStreamId;
    uiStreamId = aStreamDict.At(static_cast<C$PrjName$Application*>(Application())->AppUiUid());
    (static_cast<C$PrjName$AppUi*>(iAppUi))->RestoreL(aStore, uiStreamId);""" ) )
        appui.add_public( Method( ('TStreamId', 'Returns the id of the stream containing the data'),
                                  ('StoreL', 'Store the user interface settings to a store'),
                                  [Arg('CStreamStore&', 'aStore', 'the store to save the data to')], """\
    RStoreWriteStream stream;
    TStreamId id = stream.CreateLC(aStore); // open the stream
    iAppView->ExternalizeL(stream);         // write the data to the stream
    stream.CommitL();                       // commit the stream to file
    CleanupStack::PopAndDestroy();          // get rid of the stream
    return id;                              // Return the stream ID""", is_const = True ) )
        appui.add_public( Method( 'void', ('RestoreL', 'Restore the user interface settings from a store'),
                                  [Arg('const CStreamStore&', 'aStore', 'the store to save the data to'),
                                   Arg('TStreamId', 'aStreamId', 'the id of the stream containing the data')], """\
    RStoreReadStream stream; 
    stream.OpenLC(aStore,aStreamId); //  open the stream
    iAppView->InternalizeL(stream);  //  read in the data
    CleanupStack::PopAndDestroy();   //  get rid of the stream""" ) )
        appview.add_public( Method( 'void', ('ExternalizeL', 'Save the view settings to a stream'),
                                    [Arg('RWriteStream&', 'aStream', 'the stream to save the data to')], '',
                                    is_const = True ) )
        appview.add_public( Method( 'void', ('InternalizeL', 'Save the view settings from a stream'),
                                    [Arg('RReadStream&', 'aStream', 'the stream to save the data from')], '' ) )
        engine.add_public( Method( ('TStreamId', 'Returns the id of the store within the stream'),
                                   ('StoreL', 'Store the instance to the specified store'),
                                   [Arg('CStreamStore&', 'aStore', 'the store to stream the object to')], """\
    RStoreWriteStream stream;
    TStreamId id = stream.CreateLC(aStore); // Create a stream within the store
    ExternalizeL(stream);                   // Write the data to the stream
    stream.CommitL();
    CleanupStack::PopAndDestroy();          // get rid of the stream
    return id;                              // return the stream ID""" ) )
        engine.add_public( Method( 'void',
                                   ('RestoreL', 'Restore the instance from a stream within a store' ),
                                   [Arg('const CStreamStore&', 'aStore', 'the store containing the data'),
                                    Arg('const TStreamId&', 'aStreamId', 'the id of the stream within the store')], """\
    Clear();
    RStoreReadStream stream;
    stream.OpenLC(aStore, aStreamId);  // Open the stream
    InternalizeL(stream);              // Read the data from the stream
    CleanupStack::PopAndDestroy();     // the stream""" ) )
        engine.add_public( Method( 'void',
                                   ('ExternalizeL', 'Store the instance to the specified write stream' ),
                                   [Arg('RWriteStream&', 'aStream', 'the stream to write the object to')],
                                   '', is_const = True ) )
        engine.add_public( Method( 'void',
                                   ('InternalizeL', 'Restore the instance from the specified stream' ),
                                   [Arg('RReadStream&', 'aStream', 'the stream to read the object from')], '' ) )


def using_engine():
    """True if the init-file is using a separate engine"""
    return bool(ENG_UID)


def use_GLES( main_class, frameinterval, need_z = False, use_pixmap = False ):
    """use_GLES( main_class, frameinterval, need_z = False )
    main_class    = string_name | (string_name, string_comment)
    frameinterval = int
    need_z        = boolean
    use_pixmap    = boolean
    
    Adds variables and EGL calls to appview that initialize the drawing display, surface, and context.
    Adds also a CPeriodic that calls the rendering routine in the OpenGL ES engine.
    Creates a class for the OpenGL ES engine which contains three methods:
    AppInit() initializes the OpenGL ES state (some initializations put there already).
    AppCycle( iFrame ) performs the rendering from frame aFrame (integer).
    AppExit() is called in the end to release memory allocated in AppInit().
    """
    
    global GLESeng
    resource.libraries += [ 'libgles_cm', 'ws32' ]
    GLESeng = CClass( main_class )
    GLESeng.parent = ': public CBase'
    GLESeng.inc_inc += [ '<e32base.h> // for CBase definition',
                         '<GLES/gl.h> // OpenGL ES header file' ]
    GLESeng.two_phase( [Arg('TUint','aWidth'), Arg('TUint','aHeight')],
                      'iScreenWidth( aWidth ), iScreenHeight( aHeight )'
                      '', '' )
    GLESeng.add_private( Var( 'TUint',          'iScreenWidth', 'Width of the screen' ) )
    GLESeng.add_private( Var( 'TUint',          'iScreenHeight', 'Height of the screen' ) )
    GLESeng.add_public( Method( 'void', ('AppInit', 'Initializes the rendering object.'),
                               [], """\
    // Set the screen background color. 
    glClearColor( 0.f, 0.f, 0.f, 1.f );

    // Enable back face culling. 
    glEnable( GL_CULL_FACE  );

    // Initialize viewport and projection. 
    glViewport( 0, 0, iScreenWidth, iScreenHeight );
    glMatrixMode( GL_PROJECTION );
    glFrustumf( -1.f, 1.f, -1.f, 1.f, 3.f, 1000.f );
    glMatrixMode( GL_MODELVIEW );

    // Enable vertex arrays. 
    glEnableClientState( GL_VERTEX_ARRAY );

    // Do not use perspective correction 
    glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );""" ) )
    GLESeng.add_public( Method( 'void', ('AppExit', 'Finalizes the object.'), [], '' ) )
    GLESeng.add_public( Method( 'void', ('AppCycle', 'Renders one frame in an animation.'),
                                [Arg('TInt','aFrame','The number of the frame to be rendered.')], '' ) )
    appui.full_screen()
    if type( main_class ) == type( '' ):
        name = main_class
    else:
        name = main_class[0]

    appview.members['~$CLASS$'].code += """
        // AppExit call is made to release any allocations made in AppInit.
        if ( i%(name)s )
            {
            i%(name)s->AppExit();
            delete i%(name)s;
            }

        // release current context        
        eglMakeCurrent( iEglDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT );
        // destroy the rendering surface
        eglDestroySurface( iEglDisplay, iEglSurface );
        // destroy the rendering context
        eglDestroyContext( iEglDisplay, iEglContext );
        // release the resources of the display
        eglTerminate( iEglDisplay );""" % locals()

        
    appview.add_public( Var( 'TInt', 'iFrame', 'Frame counter variable.' ) )
    appview.add_public( Var( 'C%s *' % name, 'i%s' % name, 'An instance of class %s.' % name ) )
    appview.add_private( Var( 'EGLDisplay', 'iEglDisplay', 'The display where the graphics are drawn.' ) )
    appview.add_private( Var( 'EGLContext', 'iEglContext', 'The rendering context.' ) )
    appview.add_private( Var( 'EGLSurface', 'iEglSurface', 'The window where the graphics are drawn.' ) )
    appview.add_private( Var( 'CPeriodic*', 'iPeriodic',   'Active object for animation.' ), destroy_var = True )
    appview.add_private( Var( 'TUint',      'iTime',       'Timer tick.' ) )
    appview.add_private( Var( 'TInt',       'iFpsStartFrame',  'Calculate avrg fps of 10-19 last frames' ) )
    appview.add_private( Var( 'TUint',      'iFpsStartTime',   'Calculate avrg fps of 10-19 last frames' ) )
    appview.add_private( Var( 'TInt',       'iFpsStartFrame2', 'Calculate avrg fps of 10-19 last frames' ) )
    appview.add_private( Var( 'TUint',      'iFpsStartTime2',  'Calculate avrg fps of 10-19 last frames' ) )
    appview.inc_inc += [ '"%s.h"' % name, '<GLES/egl.h>' ]

    # set the attributes here
    attrib_list = [ 'EGL_BUFFER_SIZE', 'BufferSize' ]
    if need_z:
        # ask also for a 16bit zbuffer
        attrib_list += [ 'EGL_DEPTH_SIZE', '16' ]
    if use_pixmap:
        # ask also for a 16bit zbuffer
        attrib_list += [ 'EGL_SURFACE_TYPE', 'EGL_PIXMAP_BIT' ]
    attrib_list += [ 'EGL_NONE' ]
    attrib_list = ', '.join( attrib_list )

    if use_pixmap:
        create_surf = """
        // Create a Symbian bitmap where the graphics are blitted 
        iPixmap = new CWsBitmap( iEikonEnv->WsSession() );
        iPixmap->Create( Window().Size(), DMode);
        iEglSurface = eglCreatePixmapSurface( iEglDisplay, Config, iPixmap, NULL );"""
        appview.add_private( Var( 'CWsBitmap*', 'iPixmap', 'The bitmap where the framebuffer is drawn.' ),
                             destroy_var = True )
    else:
        create_surf = """
        // Create a window where the graphics are blitted
        iEglSurface = eglCreateWindowSurface( iEglDisplay, Config, &Window(), NULL );"""

    appview.members['ConstructL'].code += """

        iFrame = 0;         // Initialize frame counter

        EGLConfig Config;   // Describes the format, type and size of the buffers for EGLSurface

        // Get the display for drawing graphics
        iEglDisplay = eglGetDisplay( EGL_DEFAULT_DISPLAY );
        if ( iEglDisplay == NULL )
            {
            _LIT(KGetDisplayFailed, "eglGetDisplay failed");
            User::Panic( KGetDisplayFailed, 0 );
            }

        // Initialize display 
        if ( eglInitialize( iEglDisplay, NULL, NULL ) == EGL_FALSE )
            {
            _LIT(KInitializeFailed, "eglInitialize failed");
            User::Panic( KInitializeFailed, 0 );
            }

        EGLConfig *configList = NULL;            // Pointer for EGLConfigs
        EGLint numOfConfigs = 0; 
        EGLint configSize   = 0;

        // Get the number of possible EGLConfigs 
        if ( eglGetConfigs( iEglDisplay, configList, configSize, &numOfConfigs ) == EGL_FALSE )
            {
            _LIT(KGetConfigsFailed, "eglGetConfigs failed");
            User::Panic( KGetConfigsFailed, 0 );
            }

        configSize = numOfConfigs;

        // Allocate memory for the configList 
        configList = (EGLConfig*) User::Alloc( sizeof(EGLConfig)*configSize );
        if ( configList == NULL )
            {
            _LIT(KConfigAllocFailed, "config alloc failed");
            User::Panic( KConfigAllocFailed, 0 );
            }

        // Define properties for the wanted EGLSurface. To get the best possible performance, choose
        // an EGLConfig with a buffersize matching the current window's display mode
        TDisplayMode DMode = Window().DisplayMode();
        TInt BufferSize = 0;

        switch ( DMode )
             {
             case(EColor4K):
                 {
                 BufferSize = 12;
                 break;
                 }
             case(EColor64K):
                 {
                 BufferSize = 16;
                 break;
                 }
             case(EColor16M):
                 {
                 BufferSize = 24;
                 break;
                 }
             case(EColor16MU):
                 {
                 BufferSize = 32;
                 break;
                 }
             default:
                 {
                 _LIT(KDModeError,       "unsupported displaymode");
                 User::Panic( KDModeError, 0 );
                 break;
                 }
             }

        // Define properties for the wanted EGLSurface 
        const EGLint attrib_list[] = { %(attrib_list)s };

        // Choose an EGLConfig that best matches to the properties in attrib_list 
        if ( eglChooseConfig( iEglDisplay, attrib_list, configList, configSize, &numOfConfigs ) == EGL_FALSE )
            {
            _LIT(KChooseConfigFailed, "eglChooseConfig failed");
            User::Panic( KChooseConfigFailed, 0 );
            }

        // Choose the best EGLConfig. EGLConfigs returned by eglChooseConfig are sorted so 
        // that the best matching EGLConfig is first in the list.
        Config = configList[0];
        User::Free( configList ); // Free configList, not used anymore.

        %(create_surf)s
        if ( iEglSurface == NULL )
            {
            _LIT(KCreateSurfaceFailed, "creation of surface failed");
            User::Panic( KCreateSurfaceFailed, 0 );
            }

        // Create a rendering context 
        iEglContext = eglCreateContext( iEglDisplay, Config, EGL_NO_CONTEXT, NULL );
        if ( iEglContext == NULL )
            {
            _LIT(KCreateContextFailed, "eglCreateContext failed");
            User::Panic( KCreateContextFailed, 0 );
            }

        // Make the context current. Binds context to the current rendering thread and surface.
        if ( eglMakeCurrent( iEglDisplay, iEglSurface, iEglSurface, iEglContext ) == EGL_FALSE )
            {
            _LIT(KMakeCurrentFailed, "eglMakeCurrent failed");
            User::Panic( KMakeCurrentFailed, 0 );
            }

        TSize size;
        size = this->Size();

        i%(name)s = C%(name)s::NewL( size.iWidth, size.iHeight ); // Create an instance of %(name)s
        i%(name)s->AppInit();                                       // Initialize OpenGL ES

        iPeriodic = CPeriodic::NewL( CActive::EPriorityIdle );        // Create an active object for animating the scene
        iPeriodic->Start( %(frameinterval)d, %(frameinterval)d, TCallBack( $CLASS$::DrawCallBack, this ) );

        /* Get the current timer tick */
        iTime = User::TickCount();""" % locals()

    appview.add_public( Method( 'static TInt', ('DrawCallBack', 'Animates the objects, called by CPeriodic.'),
                                [ Arg( 'TAny*', 'aInstance', 'The instance the static function operates on' ) ], '' ) )

    if use_pixmap:
        swap = """
        // eglSwapBuffers is not called when using Pixmap surfaces. 
        // Instead, standard Symbian BitBlt (inside Draw()) has to be used.
        instance->DrawNow( );"""
    else:
        swap = """
        // Call eglSwapBuffers, which blits the graphics to the window 
        eglSwapBuffers( instance->iEglDisplay, instance->iEglSurface );"""
        

    appview.members['DrawCallBack'].code += """
        $CLASS$* instance = ($CLASS$*) aInstance;

        // update the frame / time counters for calculating fps
        if ( instance->iFrame == 0 )
        {
            instance->iFpsStartFrame = instance->iFrame;
            instance->iFpsStartTime  = instance->iTime;
        }
        else if ( instance->iFrame == 10 )
        {
            instance->iFpsStartFrame2 = instance->iFrame;
            instance->iFpsStartTime2  = instance->iTime;
        }
        else if ( instance->iFrame > 10 )
        {
            if ( instance->iFrame % 10 == 0 )
            {
                instance->iFpsStartFrame  = instance->iFpsStartFrame2;
                instance->iFpsStartTime   = instance->iFpsStartTime2;
                instance->iFpsStartFrame2 = instance->iFrame;
                instance->iFpsStartTime2  = instance->iTime;
            }
        }
        
        instance->iFrame++;
        // Get the current timer tick
        instance->iTime = User::TickCount();""" + """

        // Call the main OpenGL ES Symbian rendering 'loop' 
        instance->i%(name)s->AppCycle( instance->iFrame );
        %(swap)s""" % locals() + """

        // To keep the background light on 
        if ( instance->iFrame % 100 == 0 )
            {
            User::ResetInactivityTime();
            }
            
        // Suspend the current thread for a short while. Gives some time to other threads and AOs,
        // avoids the ViewSrv error in ARMI and THUMB release builds. One may try to decrease rate
        // of calling the callback function instead of this.
        if ( instance->iFrame % 50 == 0 )
            {
            User::After(0);
            }

        return 0;"""

    appview.members['Draw'].code = """
        CWindowGc& gc = SystemGc();           // Get the standard graphics context"""
    if use_pixmap:
        appview.members['Draw'].code += """
        gc.BitBlt( TPoint( 0, 0 ), iPixmap ); // Copy OpenGL's frame buffer content to screen """


__infonote_counter__ = 0

def infonote( message ):
    global __infonote_counter__
    __infonote_counter__ += 1
    c = __infonote_counter__
    return """\
            _LIT( KinMSG%d, "%s" );
            CAknInformationNote* informationNote%d = new (ELeave) CAknInformationNote;
            informationNote%d->ExecuteLD( KinMSG%d );""" % (c, message, c, c, c)

def infonote_RES( res_str ):
    resource.libraries += [ 'commonengine' ] # needed for the string loader
    appui.inc_src += [ '<stringloader.h>', '<$PrjName$.rsg>' ]
    global __infonote_counter__
    __infonote_counter__ += 1
    c = __infonote_counter__
    return """\
            // load a string from the resource file and display it
            HBufC* textResource%d = StringLoader::LoadLC( %s );
            CAknInformationNote* infoNote%d = new (ELeave) CAknInformationNote;
            infoNote%d->ExecuteLD(textResource%d->Des());
            CleanupStack::PopAndDestroy(textResource%d);""" % (c, res_str.upper(), c, c, c, c)



class Resource :
    """A class for handling the resource, mmp, etc. files in a Symbian project"""
    menu_counter = 1
    cmd_counter  = 1

    def __init__( self ):
        self.menu      = []                   # menu structure
        self.strings   = []                   # resource strings         
        self.caption   = []                   # caption for the app in application grid
        self.libraries = [ 'euser',           # the libraries needed to link the the project
                           'apparc', 'cone',
                           'eikcore', 'avkon' ]
        self.userincl  = [ r'..\inc' ]        # list of paths to search user include files
        self.extra_src = []                   # list of extra source files; entries: ( 'path', [ src, src, ... ] )
        self.exports   = []                   # if files must be exported (bld.inf, *.pkg); entries ('path', 'file)
        self.model     = 'v2.0'               # index to dictionary of supported models / SDKs
        self.multibm   = []                   # list of [name, sourcedir, type_file_list]'s
        self.groupname = ''                   # the name of the folder the app should go to
        # these things are only used if there's a separated engine apart from the Gui stuff
        self.eng_srcs  = []                   # source files for the engine
        self.eng_libs  = []                   # libraries needed to link the engine
        # currently used only by symwiz, not *.init
        self.sources   = []                   # additional source files
        self.copyfiles = []                   # src, dest pairs of files that need to be copied to create the project
        # internal attributes, don't modify outside this class (e.g., init file)
        self.classes   = []                   # which CClass objects this project has
        self.cascades  = []                   # used in creating menus
        self.default_document = ''            # name of the default document file for persistent apps
        self.supportID = {
            '7610' : ';Supports Nokia 7610\n(0x101FD5DB), 0, 0, 0, {"7610ProductID"}',
            'v2.0' : ';Supports Series 60 v2.0\n(0x101F7960), 0, 0, 0, {"Series60ProductID"}'
            }
        self.event_command_pairs = []
        self.cmd_enums = []

    def rss( self ):
        return '\n'.join( [
            'NAME RF00 // 4-letter ID, unique resource identifier for the application',
            '',
            '#include <eikon.rh>',
            '#include <eikon.rsg>',
            '#include <avkon.rh>',
            '#include <avkon.rsg>',
            '#include "$PrjName$.hrh"',
            '',
            'RESOURCE RSS_SIGNATURE { } // Needs to be left empty',
            '',
            'RESOURCE TBUF r_default_document_name { buf="%s"; }' % self.default_document,
            '',
            'RESOURCE EIK_APP_INFO  // set the default menu and CBA key',
            '    {',
            '    menubar = r_menubar_1;',
            '    cba = R_AVKON_SOFTKEYS_OPTIONS_EXIT;',
            '    }',
            '',
            self.menu_bar() ] + [
            'RESOURCE TBUF32 %s { buf="%s"; }' % tuple(s) for s in self.strings] + [
            '' ] )

    def menu_bar( self ):
        barname = 'r_menubar_%d' % Resource.menu_counter
        Resource.menu_counter += 1
        panename = barname + '_pane'
        text = 'RESOURCE MENU_BAR %s { titles = { MENU_TITLE { menu_pane = %s; } }; }\n\n' % (barname, panename)
        text += self.menu_pane( panename, self.menu )
        while self.cascades:
            c = self.cascades[0]
            self.cascades = self.cascades[1:]
            text += self.menu_pane( c[0], c[1] )
        return text
            
    def menu_pane( self, name, itemlist ):
        """Create a string for a menu pane with name, and a list of text - code pairs"""
        # figure out whether there are cascade menus
        # the items in the item list are pairs, the first element is the name
        # if the second element is a list, we need cascades
        # if it is a string, it's just a command (need to create command token for *.hrh)
        textlist = []
        for item in itemlist:
            if type( [] ) == type( item[1] ):
                # second element is a list: it's a cascade!
                # is the first element a tuple (or just a string)?
                if type( item[0] ) == type( (1,) ):
                    # yes, get the text and resource name
                    text, cascade_name = item[0]
                    assert type( '' ) == type( text )
                    assert type( '' ) == type( cascade_name )
                else:
                    # no, get the text, generate resource name
                    text = item[0]
                    assert type( '' ) == type( text )
                    cascade_name = 'r_cascade_%d' % Resource.menu_counter
                    Resource.menu_counter += 1
                textlist.append(
                    '\n'.join( [
                    '        MENU_ITEM',
                    '            {',
                    '            txt = "%s";' % text,
                    '            cascade = %s;' % cascade_name,
                    '            }' ] ) )
                self.cascades.append( ( cascade_name, item[1] ) )
            else:
                # second element is source code (text): it's a direct command
                assert type( '' ) == type( item[1] )
                # is the first element a tuple (or just a string)?
                if type( item[0] ) == type( (1,) ):
                    # yes, get the text and command name
                    text, cmdEnum = item[0]
                    assert type( '' ) == type( text )
                    assert type( '' ) == type( cmdEnum )
                else:
                    # no, get the text, generate command name
                    text = item[0]
                    assert type( '' ) == type( text )
                    cmdEnum = ('Ecmd%d_' % Resource.cmd_counter) + '_'.join(item[0].split())
                    cmdEnum = re.sub(r'\W', '', cmdEnum)
                    Resource.cmd_counter += 1
                self.cmd_enums.append( cmdEnum )
                self.event_command_pairs.append( [cmdEnum, item[1]] )
                textlist.append(
                    '\n'.join( [
                    '        MENU_ITEM',
                    '            {',
                    '            txt = "%s";' % text,
                    '            command = %s;' % cmdEnum,
                    '            }' ] ) )
        return '\n'.join( [
            'RESOURCE MENU_PANE %s' % name,
            '    {',
            '    items =',
            '        {',
            ',\n'.join(textlist),
            '        };',
            '    }',
            '' ] )
    
    def hrh( self ):
        return '\n'.join(
            [ '#ifndef __$PRJNAME$_HRH__',
              '#define __$PRJNAME$_HRH__',
              '',
              '// enumerate command codes',
              'enum T$PrjName$Ids',
              '    {',
              '    dummy = 0,',
              '    ' + ',\n    '.join( self.cmd_enums ),
              '    };',
              '',
              '#endif // __$PRJNAME$_HRH__',
              '' ] )

    def pan( self ):
        return '\n'.join(
            [ '#ifndef __$PRJNAME$_PAN__',
              '#define __$PRJNAME$_PAN__',
              '',
              '/** $PrjName$ application panic codes */',
              'enum T$PrjName$Panics ',
              '    {',
              '    E$PrjName$Ui = 1',
              '    // add further panics here',
              '    };',
              '',
              'inline void Panic(T$PrjName$Panics aReason)',
              '    {',
              '    _LIT(applicationName,"$PrjName$");',
              '    User::Panic(applicationName, aReason);',
              '    }',
              '',
              '#endif // __$PRJNAME$_PAN__',
              '', ] )

    def images( self, imgs, srcdir ):
        """copy files from srcdir to bitmap directory and export the files to emulator and phone"""
        # copy files to bitmap folder
        for i in imgs:    
            self.copyfiles.append( [join( srcdir, i ), join( PrjName, join( 'bitmaps', i ) ) ] )
            # make sure they are also copied to right location and included in the sis file
            self.exports += [ ('..\\bitmaps\\', i) ]

    def datafiles( self, datafiles, srcdir ):
        """copy files from srcdir to data directory and export the files to emulator and phone"""
        # copy files to data folder
        for d in datafiles:    
            self.copyfiles.append( [join( srcdir, d ), join( PrjName, join( 'data', d ) ) ] )
            # make sure they are also copied to right location and included in the sis file
            self.exports += [ ('..\\data\\', d) ]

    def create_mmp( self ):
        if self.caption:
            # add caption (program display name in application launcher, etc.) if there's one defined
            caption = 'RESOURCE          $PrjName$_caption.RSS\n'
        else:
            caption = '\n'
        mmpstr = '\n'.join( [ r'TARGET            $PrjName$.app',
                              r'TARGETTYPE        app',
                              r'UID               0x100039CE $UID$',
                              r'TARGETPATH        \system\apps\$prjname$',
                              r'',
                              r'SOURCEPATH        ..\data',
                              r'RESOURCE          $PrjName$.rss',
                              caption,
                              r'SYSTEMINCLUDE     \epoc32\include' ] + 
                            [ 'USERINCLUDE       %s' % i for i in self.userincl ] +
                            [ '\nSOURCEPATH        ..\src' ] +
                            [ 'SOURCE            %s' % s for s in self.sources ] +
                            [ '' ] )
        for path, srcs in self.extra_src:
            mmpstr += '\nSOURCEPATH        %s\n' % path
            for s in srcs:
                mmpstr += 'SOURCE            %s\n' % s
        # add mbm's
        for mbm in self.multibm:
            name, sourcedir, type_file_list = tuple( mbm )
            text = '\n'.join( [r'START BITMAP      %s.mbm' % name,
                               r'HEADER',
                               r'TARGETPATH        \system\apps\$PrjName$',
                               r'SOURCEPATH        ..\bitmaps'] )
            for tf in type_file_list:
                t,f = tuple(tf)
                text += '\nSOURCE            %s %s' % (t,f)
                # copy files to bitmap folder
                self.copyfiles.append( [join( sourcedir, f ), join( PrjName, join( 'bitmaps', f ) ) ] )
            mmpstr += ('\n' + text + '\nEND\n')
        # add libraries
        if using_engine():
            self.libraries += [ '$EngName$' ]
        mmpstr += '\n' + ''.join( ['LIBRARY           ' + lib + '.lib\n' for lib in self.libraries] )
        mmpstr += '\n'.join( [ r'',
                               r'LANG SC',
                               r'',
                               r'AIF               $PrjName$.aif ..\aif $PrjName$Aif.rss',
                               ] )
        if using_engine():
            mmpstr += '\nEXPORTUNFROZEN\n'
        writefile( 'group', PrjName+'.mmp', mmpstr )
            
        if using_engine():  # from here on, we're making the mmp file for the engine
            mmpstr = '\n'.join( [r'TARGET            %s.dll' % EngName,
                                 r'TARGETTYPE        dll',
                                 r'UID               0x1000008d $ENG_UID$',
                                 r'',
                                 r'USERINCLUDE       ..\inc',
                                 r'SYSTEMINCLUDE     \epoc32\include',
                                 r'',
                                 r'SOURCEPATH        ..\engine' ] +
                                [ 'SOURCE            %s' % s for s in self.eng_srcs ] +
                                [ 'LIBRARY           ' + lib + '.lib\n' for lib in self.eng_libs ] +
                                [ 'EXPORTUNFROZEN\n' ] )
            writefile( 'engine', EngName+'.mmp', mmpstr )

    def create_pkg( self ):
        pkgstr = """\
; $PrjName$.pkg
;
;Language - standard language definitions
&EN

; standard SIS file header
#{"$PrjName$"},($UID$),1,0,0

""" + self.supportID[ self.model ] + r"""

;
"SDKROOT\epoc32\release\armi\urel\$PrjName$.APP"-"!:\system\apps\$PrjName$\$PrjName$.app"
"SDKROOT\epoc32\data\z\system\apps\$PrjName$\$PrjName$.rsc"-"!:\system\apps\$PrjName$\$PrjName$.rsc"
"""
        
        if self.caption:
            pkgstr += r"""
"SDKROOT\epoc32\data\z\system\apps\$PrjName$\$PrjName$_caption.rsc"-"!:\system\apps\$PrjName$\$PrjName$_caption.rsc"
"""
        # are there extra files that must be exported to sis file?
        if self.exports:
            for path, file in self.exports:
                #pkgstr += (r'"%s%s"-"!:\system\apps\$PrjName$\%s"' % (path, file, file)) + '\n'
                pkgstr += (r'"%s%s"-"c:\system\apps\$PrjName$\%s"' % (path, file, file)) + '\n'
        # also multibitmaps need to be exported to sis file
        for mbm in self.multibm:
            name, sourcedir, type_file_list = tuple( mbm )
            pkgstr += r"""
"SDKROOT\epoc32\data\z\system\apps\$PrjName$\%s.mbm"-"!:\system\apps\$PrjName$\%s.rsc"
""" % (name, name)
        pkgstr = pkgstr.replace( 'SDKROOT', preferences.sdkroot )
        # finally, write it all to the package file
        writefile( 'sis', PrjName+'.pkg', pkgstr, add_header = False )
        
    def create_files( self ):
        # create a simple makefile
        writefile( 'group', 'Makefile', '\n'.join(
            [ '# Parse current path + EPOCROOT variable to form whole cygwin path',
              r'EPOCPATH=/$(firstword $(subst /, ,$(PWD)))$(subst \,/,$(EPOCROOT))',
              '',
              'build: emulator',
              '',
              'emulator: $PrjName$.mmp',
              '\tbldmake.bat bldfiles',
              '\tABLD.BAT build wins udeb',
              '\tabld.bat makefile %s' % preferences.ide,
              '    ',
              'sis: $PrjName$.mmp ../sis/$PrjName$.pkg',
              '\tbldmake.bat bldfiles',
              '\tABLD.BAT build armi urel',
              "\tmakesis.exe '-d$(EPOCROOT)' ../sis/$PrjName$.pkg",
              '',
              'clean:',
              '\trm -f ABLD.BAT *.DSP *.DSW *.MAKE *UID.CPP *.ncb *.plg *.OPT',
              '' ] ), 0 )

        # create *.rss, *.hrh, *.pan
        writefile( 'data', PrjName+'.rss', self.rss() )
        writefile( 'inc',   PrjName+'.hrh', self.hrh() )
        writefile( 'inc',   PrjName+'.pan', self.pan() )
        # create bld.inf
        bldinf = 'PRJ_MMPFILES\n\n'
        if using_engine():
            bldinf += '..\\engine\\%s.mmp\n' % EngName
        bldinf += '$PrjName$.mmp\n'
        if self.exports:
            bldinf += '\nPRJ_EXPORTS\n\n'
            for path, file in self.exports:
                bldinf += path + file + ' ..\\wins\\c\\system\\apps\\$prjname$\\' + file + '\n'
                bldinf += path + file + ' ..\\winscw\\c\\system\\apps\\$prjname$\\' + file + '\n'
        writefile( 'group',   'bld.inf', bldinf )
        # create rss for caption if needed
        if self.caption:
            cap, sh_cap = tuple( self.caption )
            writefile( 'data', PrjName+'_caption.RSS', '\n'.join(
                [ '#include <apcaptionfile.rh>',
                  'RESOURCE CAPTION_DATA',
                  '    {',
                  '    caption="%s";' % cap,
                  '    shortcaption="%s";' % sh_cap,
                  '    }',
                  '' ] ) )
        # create all the *.cpp and *.h files
        for c in self.classes:
            c.create_files()
        # create mmp file(s)
        self.create_mmp()
        # create pkg file
        self.create_pkg()
        # create aif .rss file
        text = [ '#include <aiftool.rh>',
                 '',
                 'RESOURCE AIF_DATA',
                 '    {',
                 '    //',
                 '    app_uid=$UID$;',
                 '    num_icons=0;',
                 '    hidden=KAppNotHidden;',
                 '    embeddability=KAppNotEmbeddable;',
                 '    newfile=KAppDoesNotSupportNewFile;',
                 '    }',
                 '' ]
        if self.groupname:
            text[-2:-2] = [ '    groupname="%s";' % self.groupname ]
        writefile( 'aif', PrjName+'Aif.rss', '\n'.join( text ) )
        # copy files if necessary
        for cf in self.copyfiles:
            shutil.copy( cf[0], cf[1] )


class CClass :
    """Class to store a Symbian class description, and to create *.cpp and *.h files from the description"""

    def __init__( self, name_and_discussion, classtype = 'C' ):
        # classtype should be C, T, M, etc., per Symbian conventions
        assert len( classtype ) == 1 
        # the first argument can be either a string (name)
        # or a tuple or list of two strings (name and class documentation / discussion)
        if type( name_and_discussion ) == type( '' ):
            self.name       = classtype + name_and_discussion
            self.discussion = ''
        else:
            assert type( name_and_discussion ) == type( ('','') )
            self.name       = classtype + name_and_discussion[0]
            self.discussion = name_and_discussion[1]
        # add this class to the resource object so that project files, etc., can be created
        resource.classes.append( self )
        self.parent     = ''           # the parent of this class
        self.inc_inc    = []           # files to be #included in the *.h file
        self.inc_src    = []           # files to be #included in the *.cpp file
        self.intro_inc  = ''           # in *.h, statements after includes and before class (forward references, etc.)
        self.intro_src  = ''           # in *.cpp, statements after includes and before functions (global variables, etc.)
        self.public     = []           # list of public members of this class
        self.protected  = []           # list of protected members of this class
        self.private    = []           # list of private members of this class
        self.members    = {}           # a dictionary of members (methods, variable, enums, etc., see method.py)
        self.namespace  = ''           # if the class defines a namespace, its name
        self.namespace_consts = ''     # consts, enums, etc., defined within the namespace
        self.is_dll     = False        # whether this class should be put into a library (dll) rather than app
        # add an empty destructor
        self.add_public( Method( '', ('~$CLASS$', ''), [], '' ) )
        

    def two_phase( self, args, ph1_init = '', ph1_code = '', ph2_code = '' ):
        """Creates a two-phase constructor for the class"""
        argstr = ', '.join( [arg.corename() for arg in args ] )
        self.add_public( Method( ('static $CLASS$*', 'A pointer to the created instance of $CLASS$'),
                                 ('NewL', 'Construct a $CLASS$ using two phase construction'), args, """\
    $CLASS$* self = NewLC( %s );
    CleanupStack::Pop( self );
    return self;""" % argstr ) )
        self.add_public( Method( ('static $CLASS$*', 'A pointer to the created instance of $CLASS$'),
                                 ('NewLC', 'Construct a $CLASS$ using two phase construction'), args, """\
    // Symbian 2-phase constructor.
    // Calls both the default C++ constructor and Symbian ConstructL methods.
    $CLASS$* self = new (ELeave) $CLASS$( %s );
    CleanupStack::PushL( self );
    self->ConstructL( %s );
    return self;""" % (argstr, argstr) ) )
        self.add_private( Method( '', ('$CLASS$', "The constructor, called by NewLC, don't use directly."),
                                  args, ph1_code, ph1_init ) )
        self.add_private( Method( 'void',
                                  ('ConstructL', 'Perform the second phase construction of a $CLASS$ object'),
                                  args, ph2_code ) )

    def make_active( self, args, ph1_init, ph1_code, ph2_code, runl_code, docancel_code ):
        # make the class to derive from CActive
        self.parent = ': public CActive'
        # remember to construct the CActive object
        if not ph1_init:
            ph1_init = 'CActive( CActive::EPriorityStandard )'
        else:
            ph1_init += ', CActive( CActive::EPriorityStandard )'
        # add to active scheduler
        ph2_code += '\nCActiveScheduler::Add( this );  // add this active object to scheduler'
        # create two-phase constructor
        self.two_phase( args, ph1_init, ph1_code, ph2_code )
        # add RunL and DoCancel methods
        self.add_private( Method( 'void',
                                  ( 'RunL',
                                    'The function that is called after the service has been completed. From CActive.' ),
                                  [], runl_code ) )
        self.add_private( Method( 'void',
                                  ( 'DoCancel',
                                    'Perform activities needed to cancel this active object. From CActive.' ),
                                  [], docancel_code ) )

    def get_members( self, memberlist ):
        enums, methods, vars = [],[],[]
        for m in memberlist:
            if m.__class__ == Enum:   enums.append( m )
            elif m.__class__ == Var:  vars.append( m )
            else:                     methods.append( m )
        return enums + methods + vars

    def add_public( self, member, destroy_var = False, destroy_arr = False ):
        member.is_dll = self.is_dll
        self.members[ member.name ] = member
        self.public.append( member )
        if destroy_var:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete %s; %s = NULL;' % (member.name, member.name)
        if destroy_arr:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete[] %s; %s = NULL;' % (member.name, member.name)
            
    def add_protected( self, member, destroy_var = False, destroy_arr = False ):
        member.is_dll = self.is_dll
        self.members[ member.name ] = member
        self.protected.append( member )
        if destroy_var:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete %s; %s = NULL;' % (member.name, member.name)
        if destroy_arr:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete[] %s; %s = NULL;' % (member.name, member.name)

    def add_private( self, member, destroy_var = False, destroy_arr = False ):
        member.is_dll = self.is_dll
        self.members[ member.name ] = member
        self.private.append( member )
        if destroy_var:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete %s; %s = NULL;' % (member.name, member.name)
        if destroy_arr:
            # this is a variable, owned by this object, destroy in destructor
            self.members['~$CLASS$'].code += '\n    delete[] %s; %s = NULL;' % (member.name, member.name)

    def add_notify( self ):
        """Add 3 methods, one for getting the error in text form, one for showing errors,
        one for showing information notifys"""
        errors = [ #'KErrNotFound', 'KErrNoMemory', 'KErrInUse', 'KErrNotReady', 'KErrNotSupported' ]
            'KErrNotFound',
            'KErrGeneral',
            'KErrCancel',
            'KErrNoMemory',
            'KErrNotSupported',
            'KErrArgument',
            'KErrTotalLossOfPrecision',
            'KErrBadHandle',
            'KErrOverflow',
            'KErrUnderflow',
            'KErrAlreadyExists',
            'KErrPathNotFound',
            'KErrDied',
            'KErrInUse',
            'KErrServerTerminated',
            'KErrServerBusy',
            'KErrCompletion',
            'KErrNotReady',
            'KErrUnknown',
            'KErrCorrupt',
            'KErrAccessDenied',
            'KErrLocked',
            'KErrWrite',
            'KErrDisMounted',
            'KErrEof',
            'KErrDiskFull',
            'KErrBadDriver',
            'KErrBadName',
            'KErrCommsLineFail',
            'KErrCommsFrame',
            'KErrCommsOverrun',
            'KErrCommsParity',
            'KErrTimedOut',
            'KErrCouldNotConnect',
            'KErrCouldNotDisconnect',
            'KErrDisconnected',
            'KErrBadLibraryEntryPoint',
            'KErrBadDescriptor',
            'KErrAbort',
            'KErrTooBig',
            'KErrDivideByZero',
            'KErrBadPower',
            'KErrDirFull',
            'KErrHardwareNotAvailable',
            'KErrSessionClosed',
            'KErrPermissionDenied',
            ]
        self.inc_inc += [ '<aknnotewrappers.h>' ]
        self.add_private( Method( 'TDesC&', 'GetErrText',
                                  [ Arg('TDes&', 'aText'), Arg('TInt', 'aError') ], """\
    switch (aError)
        {
        case KErrNone:
            {
            aText.Copy(_L("OK"));
            break;
            }""" + ''.join( [ """
        case ERROR: { aText.Copy(_L("ERROR")); break; }""".replace( 'ERROR', e ) for e in errors ] ) + """
        default:
            {
            aText = _L("Unknown");
            }
        }
    return ((TDesC&)aText);\n""") )
        self.add_private( Method( 'void', 'ErrorNotifyL',
                                  [ Arg('const TDesC&', 'aErrorTxt', "A string noting which function's errors are notified."),
                                    Arg('TInt', 'aError'),
                                    Arg('TBool', 'aNotify = EFalse', 'If this is true, print also OK for no error, otherwise do nothing if no error.') ], """\
    if ((aError != KErrNone) || aNotify)
        {
        TBuf <20> errorText;
        GetErrText(errorText, aError);
        TBuf<80> text;
        text.Format(_L("%S: %S."), &aErrorTxt, &errorText);

        CAknErrorNote* errorNote = new (ELeave) CAknErrorNote;
        errorNote->ExecuteLD(text);
        }\n""") ) 
        self.add_private( Method( 'void', 'InfoNotifyL',
                                  [ Arg('const TDesC&', 'aInfoTxt') ], """\
    CAknInformationNote *InformationNote = new (ELeave) CAknInformationNote;
    InformationNote->ExecuteLD( aInfoTxt );""" ) )
        
    def use_camera( self ):
        """Add the basic machinery for using a camera"""
        self.inc_inc += [ '<ecam.h>' ]
        self.parent  += ', public MCameraObserver'
        if self.members.has_key('iAppView'):
            # if there's already a pointer to iAppview, don't repeat it
            # also, assume that constructors have already been created
            self.members['ConstructL'].code += '\n    iCamera = CCamera::NewL( *this, 0 );'
        else:
            self.two_phase( [Arg('C$PrjName$AppView*','aAppView')], 'iAppView(aAppView)', '',
                            '    iCamera = CCamera::NewL( *this, 0 );' )
            self.inc_inc += [ '"$PrjName$AppView.h"' ]
            self.add_private( Var( 'C$PrjName$AppView*', 'iAppView' ) )
        self.add_notify()
        self.members['~$CLASS$'].code += """
    iCamera->PowerOff();
    iCamera->Release();"""
        self.add_private( Var( 'CCamera*',           'iCamera' ), destroy_var = True )
        self.add_private( Method( 'void', ('ReserveComplete', 'Camera reservation ready. From MCameraObserver.'),
                                  [Arg('TInt', 'aError')], """\
    if (aError != KErrNone)
        {
        ErrorNotifyL(_L("Camera reserve err: "), aError, ETrue);
        return;
        }""" ) )
        self.add_private( Method( 'void', ('PowerOnComplete', 'Power is on. From MCameraObserver.'),
                                  [Arg('TInt', 'aError')], """\
    if (aError != KErrNone)
        {
        ErrorNotifyL(_L("Camera power err: "), aError, ETrue);
        }""" ) )
        self.add_private( Method( 'void',
                                  ('ImageReady', 'Captured image ready, now client can draw it. From MCameraObserver.'),
                           [Arg('CFbsBitmap*', 'aBitmap'),
                            Arg('HBufC8*',     'aData'),
                            Arg('TInt',        'aError'),], """\
    if (aError != KErrNone)
        {
        ErrorNotifyL(_L("Image capture err: "), aError, ETrue);
        return;
        }""" ) )
        self.add_private( Method( 'void', ('FrameBufferReady', 'Video frame buffer filled. From MCameraObserver.'),
                                  [Arg('MFrameBuffer*','aFrameBuffer'), Arg('TInt','aError')], """\
    if (aError != KErrNone)
        {
        ErrorNotifyL(_L("Frame ready err: "), aError, ETrue);
        return;
        }""" ) )
        self.add_private( Method( 'void', ('ViewFinderFrameReady',
                                           'ViewFinder frame is ready, now can draw the display. From MCameraObserver.'),
                                  [Arg('CFbsBitmap&','aFrame')], '    iAppView->DrawImage( &aFrame );' ) )
        resource.libraries += [ 'ecam' ]
        appview.add_public( Method( 'void', ('DrawImage', 'Draw the captured image on the display'),
                                    [Arg('CFbsBitmap*', 'aImage')], """\
    // Get the standard graphics context
    CWindowGc& gc = SystemGc();
    // Update the screen
    gc.Activate( Window() );
    TRect drawRect( Rect() );
    if ( aImage->SizeInPixels().iWidth > drawRect.Width() )
        {
        // do a scaled blit
        gc.DrawBitmap( drawRect, aImage );
        }
    else
        {
        // Find a position so that the bitmap image is centered on the display
        TPoint pos;
        pos.iX = ( drawRect.iBr.iX - aImage->SizeInPixels().iWidth ) / 2;
        pos.iY = ( drawRect.iBr.iY - aImage->SizeInPixels().iHeight ) / 2;
        gc.Clear( drawRect );
        gc.BitBlt( pos, aImage );
        }
    gc.Deactivate();
    iCoeEnv->WsSession().Flush();""", is_const = True ))
        self.add_private( Method( 'void', ('StartViewFinder', 'Start the view finder.'), [], """\
    TSize drawSize( 160, 120 );

    if (iCamera->ViewFinderActive())
        {
        ErrorNotifyL(_L("View Finder already active: "), KErrInUse, ETrue);
        return;
        }

    TRAPD(err, iCamera->StartViewFinderBitmapsL( drawSize ));
    ErrorNotifyL(_L("View Finder start err: "), err, EFalse);

    if (!iCamera->ViewFinderActive())
        {
        ErrorNotifyL(_L("View Finder not active: "), KErrNotReady, ETrue);
        }""" ) )
        self.add_private( Method( 'void', ('StopViewFinder', 'Stop the view finder.'), [], """\
    if (!iCamera->ViewFinderActive())
        {
        ErrorNotifyL(_L("View Finder not active: "), KErrInUse, ETrue);
        return;
        }

    TRAPD(err, iCamera->StopViewFinder());
    ErrorNotifyL(_L("View Finder stop err: "), err, EFalse);""" ) )

    def header_file( self ):
        """Create the text for the header file for this class"""
        classdoc = """
/*!
  @class %(name)s

  @discussion %(discussion)s
  */""" % self.__dict__
        # protection for not including file many times
        protect = '__' + self.name.upper() + '_H__'
        classbody  = '\n'.join( [
                          classdoc,
                          'class %(name)s %(parent)s' % self.__dict__,
                          '    {',
                          'public:\n\n'    + ''.join( [m.declaration() for m in self.get_members(self.public)] ),
                          'protected:\n\n' + ''.join( [m.declaration() for m in self.get_members(self.protected)] ),
                          'private:\n\n'   + ''.join( [m.declaration() for m in self.get_members(self.private)] ),
                          '    };',
                          ] )
        if self.namespace:
            classbody = 'namespace ' + self.namespace + '\n{\n' + self.namespace_consts + classbody + '\n}\n'
        txt = '\n'.join( ['#ifndef ' + protect,
                          '#define ' + protect,
                          '',
                          '\n'.join( ['#include %s' % s for s in self.inc_inc] ),
                          '',
                          self.intro_inc,
                          classbody,
                          '',
                          '#endif // ' + protect,
                          ''
                          ] )
        return txt.replace( '$CLASS$', self.name )

    def source_file( self ):
        """Create the text for the source file for this class"""
        includes = '\n'.join( ['#include ' + s for s in self.inc_src] ) + '\n\n'
        if self.namespace:
            includes += 'using namespace ' + self.namespace + ';\n\n'
        intro = self.intro_src + '\n\n'
        code = ''.join( [m.definition() for m in
                         self.get_members(self.public) +
                         self.get_members(self.protected) +
                         self.get_members(self.private)] )
        if EngName == self.name[1:]:
            code += """
// requirement for E32 DLLs
EXPORT_C TInt E32Dll(TDllReason)
    {
    return KErrNone;
    }
"""
        return (includes + intro + code).replace( '$CLASS$', self.name )

    def create_files( self ):
        self.add_later()
        self.name = self.name.replace('$PrjName$', PrjName)
        # file names don't have the leading 'C' or 'T', etc., of a class name
        fname = self.name[1:]
        # include the header file 
        self.inc_src[:0] = [ '"'+fname+'.h"' ]
        writefile( 'inc', fname+'.h', self.header_file() )
        src_name = fname+'.cpp'
        if self.is_dll:
            writefile( 'engine', src_name, self.source_file() )
            resource.eng_srcs.append( src_name )
        else:
            writefile( 'src', src_name, self.source_file() )
            resource.sources.append( src_name )

    def add_later( self ):
        # inherit this if you want to do part of the initialization after
        # the initfile has been processed
        pass
        

class Application( CClass ):
    """Class for creating the Application class for the project"""

    def __init__( self ):
        CClass.__init__( self, ('$PrjName$Application', 'The application part') )
        self.parent     = ': public CAknApplication'
        self.inc_inc    = [ '<aknapp.h>' ]
        self.inc_src    = [ '"$PrjName$Document.h"' ]
        
        self.add_protected( Method( 'CApaDocument*', 'CreateDocumentL', [],
                                    '    return (static_cast<CApaDocument*>(C$PrjName$Document::NewL(*this)));') )
        self.add_public( Method( 'TUid', 'AppDllUid', [], """    return KUid$PrjName$App;""", is_const = True) )
        self.intro_src  = '\n'.join(
            [
            'const TUid KUid$PrjName$App = {$UID$};'
            '',
            '// DLL entry point, return that everything is ok',
            'GLDEF_C TInt E32Dll(TDllReason /*aReason*/)',
            '    {',
            '    return KErrNone;',
            '    }',
            '',
            '// Create an application, and return a pointer to it',
            'EXPORT_C CApaApplication* NewApplication() ',
            '    {',
            '    return (static_cast<CApaApplication*>(new C$PrjName$Application));',
            '    }',
            '' ] )

class Document( CClass ):
    """Class for creating the Document class for the project"""

    def __init__( self ):
        CClass.__init__( self, ('$PrjName$Document', 'The document part') )
        #self.parent     = ': public CAknDocument'  # with CAknDocument file storage doesn't seem to work...
        self.parent     = ': public CEikDocument'
        self.inc_inc    = [ '<akndoc.h>' ]
        self.inc_src    = [ '"$PrjName$AppUi.h"', '"$PrjName$Application.h"' ]
        self.intro_inc  = 'class C$PrjName$AppUi;\nclass CEikApplication;'

        aApp = Arg('CEikApplication&', 'aApp', 'Application that creates this document')
        self.two_phase( [aApp], ['CEikDocument( aApp )'], '', '' )
        #self.two_phase( [aApp], ['CAknDocument( aApp )'], '', '' )
        self.add_public( Method( ('CEikAppUi*', 'A pointer to the created AppUi'),
                                 ('CreateAppUiL', 'Create an AppUi'), [],
                                 '    return (static_cast <CEikAppUi*> (new (ELeave) C$PrjName$AppUi( this )));' ) )


class AppUi( CClass ):
    """The AppUi class for a Symbian project"""

    def __init__( self ):
        CClass.__init__( self, ('$PrjName$AppUi', 'The user interface part') )
        self.parent     = ': public CAknAppUi'
        self.inc_inc    = [ '<aknappui.h>', '"$PrjName$AppView.h"', '"$PrjName$Document.h"' ]
        self.inc_src    = [ '<avkon.hrh>', '<aknnotewrappers.h>', '"$PrjName$.pan"', '"$PrjName$.hrh"' ]
        self.add_public( Method( 'void', 'ConstructL', [], """\
    BaseConstructL();               // initialize appUI with standard values, read standard resource files
    iAppView = C$PrjName$AppView::NewL( ClientRect(), iDocument, this ); // construct appView
    iAppView->SetMopParent( this ); // allow object to be in network of object providers (see MObjectProvider)
    AddToStackL( iAppView );        // view now gets key press events
    """ ) )
        self.members['~$CLASS$'].code += """\
    if (iAppView)
        {
        RemoveFromStack( iAppView );
        delete iAppView;  iAppView = NULL;
        }"""
        self.add_public( Method( '', '$CLASS$', [Arg('C$PrjName$Document*', 'aDocument', 'the model')],
                                 '', ['iDocument(aDocument)'] ) )
        self.add_private( Var( 'C$PrjName$AppView*', 'iAppView', 'The application view' ) )
        self.add_private( Var( 'C$PrjName$Document*', 'iDocument', 'The current document' ) )
            
    def full_screen( self ):
        cc = self.members['ConstructL']
        cc.code = cc.code.replace( 'ClientRect', 'ApplicationRect', 1 )

    def dynamic_menu( self, code ):
        self.add_private( Method( 'void',
                                  ('DynInitMenuPaneL',
                                   'From MEikMenuObserver, sets state of menu items (dimming, changing text, etc.).'),
                                  [Arg('TInt', 'aResourceId', 'Resource ID identifying the menu pane to initialise.'),
                                   Arg('CEikMenuPane*', 'aMenuPane', 'The in-memory representation of the menu bar.')],
                                  code ) )
        resource.libraries += [ 'eikcoctl' ]
        self.inc_src += [ '<$PrjName$.rsg>', '<eikmenup.h>' ]

    def command_handler( self ):
        """This function creates a command handler"""
        return """\
    switch( aCommand )
        {
        case EEikCmdExit:
        case EAknSoftkeyExit:
            {
            Exit();
            break;
            }\n""" + '\n'.join( ["""
        case %s:
            {
%s
            break;
            }""" % tuple(c) for c in resource.event_command_pairs ] ) + """

        default:
            {
            Panic( E$PrjName$Ui );
            break;
            }
        }\n"""

    def add_later( self ):
        self.add_public( Method( 'void', ('HandleCommandL', 'Handle user menu selections'),
                                 [Arg('TInt', 'aCommand')],                                         
                                 self.command_handler() ) )


class AppView( CClass ):
    """The AppView class for a Symbian project"""
    def __init__( self ):
        CClass.__init__( self, ('$PrjName$AppView', self.__class__.__doc__) )
        self.parent     = ': public CCoeControl'
        self.inc_inc    = [ '<coecntrl.h>', '"$PrjName$Document.h"' ]
        self.intro_inc  = 'class $PrjName$AppUi;'
        self.inc_src    = [ '<coemain.h>', '"$PrjName$AppUi.h"' ]

        self.add_private( Method( 'void', ('Draw', 'Draw the appview to screen'),
                                  [Arg('const TRect&', 'aRect', 'The rectangle to draw into')], """\
    CWindowGc& gc = SystemGc();    // Get the standard graphics context 
    TRect rect = Rect();           // Get the control's extent
    gc.Clear( rect );              // Clear the screen""", is_const = True ) )
        self.two_phase([Arg('const TRect&', 'aRect', ''),
                        Arg('C$PrjName$Document*', 'aDocument', ''),
                        Arg('C$PrjName$AppUi*', 'aAppUi', '') ],
                       ph1_init = 'iDocument(aDocument), iAppUi(aAppUi)',
                       ph1_code = '',
                       ph2_code = """\
    CreateWindowL();    // Create a window for this application view
    SetRect( aRect );   // Set the window size
    ActivateL();        // Activate the window, which makes it ready to be drawn""")
        self.add_private( Var( 'C$PrjName$Document*', 'iDocument', 'The document' ) )
        self.add_private( Var( 'C$PrjName$AppUi*', 'iAppUi', 'Pointer to the appui class' ) )
        self.members['~$CLASS$'].code = "    iDocument = NULL; // not the owner, so don't delete"


    def keyevents( self, icode_callback_pairs, input_cap = 'ENone' ):
        keyeventhandler = """\
    // we only want the key press, not the key up/down event
    if (aType == EEventKey)
        {""" + '\n'.join( ["""
        if( aKeyEvent.iCode == %s )
            {
%s
            DrawNow();
            return EKeyWasConsumed;
            }""" % tuple( pair ) for pair in icode_callback_pairs ] ) + """
        }
    // return the default functionality
    return CCoeControl::OfferKeyEventL(aKeyEvent, aType);"""

        self.add_public( Method( ('TKeyResponse', 'whether key was consumed'),
                                 ('OfferKeyEventL', 'handle keypresses'),
                                 [Arg( 'const TKeyEvent&', 'aKeyEvent', 'holds the event data' ),
                                  Arg( 'TEventCode', 'aType', 'type of the event' )],
                                 keyeventhandler ) )
        self.add_public( Method( ('TCoeInputCapabilities', 'the capabilities of this class'),
                                 ('InputCapabilities', 'capabilities of OfferKeyEventL'),
                                 [], """\
    return TCoeInputCapabilities::%s;""" % input_cap, is_const = True ) )
        
                        
def apply_prjname( text ):
    text = text.replace('$PRJNAME$', PRJNAME)
    text = text.replace('$PrjName$', PrjName)
    text = text.replace('$prjname$', prjname)
    return text.replace('$EngName$', EngName)
    
header = """\
/*********************************************************************
 * FNAME
 *
 * COPYRIGHT
 *********************************************************************/

"""

def writefile( subdir, fname, text, add_header = 1 ):
    dir = join( PrjName, subdir )
    f = file( join( dir, fname ), 'w' )
    if add_header:
        f.write( header.replace( 'FNAME', fname ) )
    text = text.replace('$UID$', UID)
    text = text.replace('$ENG_UID$', ENG_UID)
    text = apply_prjname( text )
    if os.linesep == '\r\n':
        text.replace( '\n', '\r\n' )
    f.write( apply_prjname( text ) )
    f.close()


def create_dirs():
    # create the directories if needed
    dirs = ['aif', 'bitmaps', 'data', 'group', 'inc', 'sis', 'src']
    if using_engine():
        dirs.append( 'engine' )
    old = PrjName + '.old'
    for d in dirs:
        od,nd = join( old, d ), join( PrjName, d )
        if not os.path.exists( nd ):
            os.makedirs( nd )
        if not os.path.exists( od ):
            os.makedirs( od )
    # create a file that is removed if the run was succesful
    failed = join( old, 'failed' )
    if not os.path.exists( failed ):
        # last run was success
        # move all the files to the save tree
        for d in dirs:
            od,nd = join( old, d ), join( PrjName, d )
            for f in os.listdir( nd ):
                shutil.copy( join( nd,f ), join( od,f ) )
        file( failed, 'w' ).close()     # put in a marker that stays if run fails


def create_diff():
    """If the project was completed succesfully, and there were differences compared to
    the previous run, find the differences and show them in an editor"""
    dirs = ['aif', 'bitmaps', 'data', 'group', 'inc', 'sis', 'src']
    if using_engine():
        dirs.append( 'engine' )
    new = PrjName
    old = new + '.old'
    fname = join( old, 'diff.txt' )
    diff_file = file( fname, 'w' )
    # compare all files in project tree to files in save tree
    for d in dirs:
        od, nd = join( old, d ), join( PrjName, d )
        for f in os.listdir( nd ):
            ndf, odf = join( nd,f ), join( od,f )
            if os.path.exists( odf ):
                diff = difflib.context_diff( file( ndf ).readlines(),
                                             file( odf ).readlines(),
                                             ndf, odf )
                diff_file.writelines( diff )
            else:
                # if the file doesn't exist in the save tree, it must be a new file
                print 'Created ' + ndf
#     for root, dirs, files in os.walk( new, topdown = False ):
#         for name in files:
#             currname = join( root, name )
#             savename = currname.replace( new, old, 1 )
#             if os.path.exists( savename ):
#                 diff = difflib.context_diff( file( currname ).readlines(),
#                                              file( savename ).readlines(),
#                                              currname, savename )
#                 diff_file.writelines( diff )
#             else:
#                 # if it's a file in any of the subdirectories, doesn't
#                 # exist in the save tree, it must be a new file
#                 if root != PrjName:
#                     print 'Created ' + currname
    diff_file.close()
    # got this far without failure, remove marker file
    failed = join( old, 'failed' )
    if os.path.exists( failed ):
        os.remove( failed )
    # if there were differences, show in an editor
    if os.path.getsize( fname):
        os.system( preferences.editor + ' ' + fname )
            

###########################
#    THE MAIN SECTION     #
###########################
if __name__ == '__main__':
    # read in the command line options, print usage help string if there are problems
    # right now only the initfile is expected
    usage = "usage: %prog [options] project"
    parser = OptionParser( usage = usage )
    parser.add_option( '-d', '--dummy',
                       action='store_true', dest='dummy', default=False,
                       help='this is just a dummy')
    (options, args) = parser.parse_args()
    
    #print options
    #print args
    initfile = args[0]

    # create the main objects, one for handling resources, then 4 different CClasses
    # for the key Symbian UI classes
    # note that resource will have a list of references to all CClasses, including these 4
    resource    = Resource( )
    application = Application()
    document    = Document()
    appui       = AppUi()
    appview     = AppView()

    # read in the initfile, execute the Python code in that file
    execfile( initfile )

    # make two more versions of the project name
    # the initfile should contain PrjName (mixed case), make lower and upper case versions
    prjname = PrjName.lower()
    PRJNAME = PrjName.upper()

    # create the directories needed, possibly save the previous version to PrjName.old
    # directory, clear the working directories
    create_dirs()

    # create all the files, the class files, and project and resource files
    resource.create_files()

    # check if there were any changes, if there were, then pop up an editor that shows those
    # differences
    create_diff()

